from . import utils
from .client import Client
from .connection import Connection
from .packet import Packet
from .player import Profile, Stats
from .tribe import *

from .__version__ import __author__, __title__, __description__
from .__version__ import __url__, __version__, __credits__, __license__